package meerkat.bulletinboard;

import com.google.common.util.concurrent.*;
import com.google.protobuf.ByteString;
import meerkat.bulletinboard.callbacks.GetRedundancyFutureCallback;
import meerkat.bulletinboard.callbacks.PostMessageFutureCallback;
import meerkat.bulletinboard.callbacks.ReadMessagesFutureCallback;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Voting.*;

import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Created by Arbel Deutsch Peled on 05-Dec-15.
 * Thread-based implementation of a Async Bulletin Board Client.
 * Features:
 * 1. Handles tasks concurrently.
 * 2. Retries submitting
 */
public class ThreadedBulletinBoardClient extends SimpleBulletinBoardClient implements AsyncBulletinBoardClient {

    private final static int THREAD_NUM = 10;
    ListeningExecutorService listeningExecutor;

    private final static int READ_MESSAGES_RETRY_NUM = 1;

    private int minAbsoluteRedundancy;

    /**
     * Stores database locations and initializes the web Client
     * Stores the required minimum redundancy.
     * Starts the Thread Pool.
     * @param clientParams contains the required information
     */
    @Override
    public void init(BulletinBoardClientParams clientParams) {

        super.init(clientParams);

        minAbsoluteRedundancy = (int) (clientParams.getMinRedundancy() * clientParams.getBulletinBoardAddressCount());

        listeningExecutor = MoreExecutors.listeningDecorator(Executors.newFixedThreadPool(THREAD_NUM));

    }

    /**
     * Post message to all DBs
     * Retry failed DBs
     * @param msg is the message,
     * @return the message ID for later retrieval
     * @throws CommunicationException
     */
    @Override
    public MessageID postMessage(BulletinBoardMessage msg, ClientCallback<?> callback){

        // Create job
        BulletinClientJob job = new BulletinClientJob(meerkatDBs, minAbsoluteRedundancy, BulletinClientJob.JobType.POST_MESSAGE, msg, -1);

        // Submit job and create callback
        Futures.addCallback(listeningExecutor.submit(new BulletinClientWorker(job)), new PostMessageFutureCallback(listeningExecutor, callback));

        // Calculate the correct message ID and return it
        digest.reset();
        digest.update(msg.getMsg());
        return MessageID.newBuilder().setID(ByteString.copyFrom(digest.digest())).build();
    }

    @Override
    public MessageID postBatch(byte[] signerId, int batchId, List<BatchData> batchDataList, int startPosition, ClientCallback<?> callback) {
        return null; //TODO: Implement
    }

    @Override
    public MessageID postBatch(byte[] signerId, int batchId, List<BatchData> batchDataList, ClientCallback<?> callback) {
        return null; //TODO: Implement
    }

    /**
     * Access each database and search for a given message ID
     * Return the number of databases in which the message was found
     * Only try once per DB
     * Ignore communication exceptions in specific databases
     */
    @Override
    public void getRedundancy(MessageID id, ClientCallback<Float> callback) {

        // Create job
        BulletinClientJob job = new BulletinClientJob(meerkatDBs, minAbsoluteRedundancy, BulletinClientJob.JobType.GET_REDUNDANCY, id, 1);

        // Submit job and create callback
        Futures.addCallback(listeningExecutor.submit(new BulletinClientWorker(job)), new GetRedundancyFutureCallback(listeningExecutor, callback));

    }

    /**
     * Go through the DBs and try to retrieve messages according to the specified filter
     * If at the operation is successful for some DB: return the results and stop iterating
     * If no operation is successful: return null (NOT blank list)
     */
    @Override
    public void readMessages(MessageFilterList filterList, ClientCallback<List<BulletinBoardMessage>> callback) {

        // Create job
        BulletinClientJob job = new BulletinClientJob(meerkatDBs, minAbsoluteRedundancy, BulletinClientJob.JobType.READ_MESSAGES,
                filterList, READ_MESSAGES_RETRY_NUM);

        // Submit job and create callback
        Futures.addCallback(listeningExecutor.submit(new BulletinClientWorker(job)), new ReadMessagesFutureCallback(listeningExecutor, callback));

    }

    @Override
    public void readBatch(byte[] signerId, int batchId, ClientCallback<SignedBatch> callback) {
        // TODO: Implement
    }

    @Override
    public void close() {
        super.close();

        try {
            listeningExecutor.shutdown();
            while (! listeningExecutor.isShutdown()) {
                listeningExecutor.awaitTermination(10, TimeUnit.SECONDS);
            }
        } catch (InterruptedException e) {
            System.err.println(e.getCause() + " " + e.getMessage());
        }
    }

}
